package dnsfilter.android.dnsserverconfig.widget;

public class NotDeserializableException extends Exception{

    public NotDeserializableException() {
        super();
    }

    public NotDeserializableException(String message) {
        super(message);
    }
}
